# Copyright: Hallux team, 2023

from __future__ import annotations

import os
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory


@contextmanager
def set_directory(path: str | Path | TemporaryDirectory):
    """Sets the cwd within the context

    Args:
        path (Path): The path to the cwd

    Yields:
        None
    """
    origin = Path().absolute()
    path_str = path.name if isinstance(path, TemporaryDirectory) else str(path)
    try:
        os.chdir(path_str)
        yield
    finally:
        os.chdir(origin)


def find_arg(argv: list[str], name: str) -> int:
    """
    Finds argument index
    """
    for i in range(len(argv)):
        arg = argv[i]
        if arg == name or arg.startswith(name):
            return i
    return -1


def find_argvalue(argv: list[str], name: str) -> str | None:
    """
    Finds argument index and following value
    """
    for i in range(len(argv)):
        arg = argv[i]
        if arg == name:  # --key value
            if len(argv) > i + 1 and not argv[i + 1].startswith("-"):
                return argv[i + 1]
        elif arg.startswith(name + "="):  # --key=value
            return arg[len(name) + 1 :]

    return None



# Copyright: Hallux team, 2023

from __future__ import annotations

import subprocess
from pathlib import Path

from ...auxilary import set_directory
from ...issues.issue import IssueDescriptor
from ..issue_solver import IssueSolver
from ..python.python_issue import PythonIssue


class Ruff_IssueSolver(IssueSolver):
    def __init__(
        self,
        config_path: Path,
        run_path: Path,
        command_dir: str = ".",
        validity_test: str | None = None,
        args: str | None = None,
    ):
        super().__init__(config_path, run_path, command_dir, validity_test=validity_test)

        self.args: str = args if args is not None else "check"

    def list_issues(self) -> list[IssueDescriptor]:
        issues: list[IssueDescriptor] = []
        with set_directory(self.run_path):
            try:
                ruff_output = subprocess.check_output(["ruff", self.args, self.command_dir])
            except subprocess.CalledProcessError as e:
                ruff_output = e.output

        issues.extend(PythonIssue.parseIssues(ruff_output.decode("utf-8")))

        return issues

    def solve_issues(self, diff_target, query_backend):
        print("Process ruff:")
        super().solve_issues(diff_target, query_backend)



# Copyright: Hallux team, 2023

from __future__ import annotations

import subprocess
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Final

from hallux.logger import logger

from ..auxilary import set_directory
from ..backends.query_backend import QueryBackend
from ..issues.issue import IssueDescriptor
from ..proposals.diff_proposal import DiffProposal
from ..targets.diff import DiffTarget


class IssueSolver(ABC):
    """
    Base abstract class for issue solving.
    For every tool such as 'ruff', or 'make', one need to inherit a class out and define list_issues()
    Issues repeatedly solved one-by-one, with check that fix was appropriate (is_issue_fixed func is used).
    If fix wasn't successful, we ignore issue (and revert corresponding fix) and go to the next one.
    """

    def __init__(
        self,
        config_path: Path,
        run_path: Path,
        command_dir: str = ".",
        validity_test: str | None = None,
    ):
        self.config_path: Final[Path] = config_path
        self.run_path: Final[Path] = run_path
        self.command_dir: Final[str] = command_dir
        self.validity_test: Final[str] = validity_test

        if validity_test is not None:
            try:
                logger.info(f"Try running validity test: {validity_test} ...")

                with set_directory(self.config_path):
                    subprocess.check_output(
                        ["bash"] + validity_test.split(" "),
                    )
                    logger.info("\033[92m PASSED\033[0m")
            except subprocess.CalledProcessError as e:
                raise SystemError(f"Validity Test '{validity_test}' is failing right from the start") from e

    @abstractmethod
    def list_issues(self) -> list[IssueDescriptor]:
        """
        Shall be implemented in child class
        :return: List of issued for a particular Processor/Solver
        """
        pass

    def is_issue_fixed(self) -> bool:
        """
        :returns: True, if latest fix was successful
        """
        if self.validity_test is not None:
            try:
                with set_directory(self.config_path):
                    subprocess.check_output(["bash"] + self.validity_test.split(" "))
                logger.info(f"validity test: {self.validity_test} PASSED")
                return True
            except subprocess.CalledProcessError:
                logger.info(f"\033[91m validity test: {self.validity_test} FAILED\033[0m")
                return False
        else:
            new_issues = self.list_issues()
            # Number of issues decreased => FIX SUCCESFULL
            return len(new_issues) < len(self.target_issues)

    def solve_issues(self, diff_target: DiffTarget, query_backend: QueryBackend):
        issue_index: int = 0
        self.target_issues = self.list_issues()
        while issue_index < len(self.target_issues):
            issue = self.target_issues[issue_index]

            proposals = issue.list_proposals()
            fixing_successful: bool = False

            proposal: DiffProposal | None = None
            for proposal in proposals:
                applying_successful: bool
                used_backend = None

                iters = 10
                while used_backend != query_backend and iters > 0:
                    iters -= 1
                    try:
                        applying_successful, used_backend = proposal.try_fixing_with_priority(
                            diff_target=diff_target, query_backend=query_backend, used_backend=used_backend
                        )
                    except Exception as e:
                        diff_target.revert_diff()
                        raise e

                    if applying_successful:
                        fixing_successful = self.is_issue_fixed()
                    else:
                        diff_target.revert_diff()
                        continue

                    if fixing_successful:
                        # this is a good place for "multi-step proposal" extension
                        fixing_successful = diff_target.commit_diff()
                        # commit_diff() might not be OK
                        if fixing_successful:
                            break

                if fixing_successful:
                    break

                # if whole loop passed, but there were no successful fix, revert the change
                diff_target.revert_diff()

            if fixing_successful:
                # provide feedback, in order to collect training data
                query_backend.report_succesfull_fix(issue, proposal)
                if diff_target.requires_refresh():
                    self.target_issues = self.list_issues()
                else:
                    issue_index += 1
                logger.message(
                    f"{issue.filename}:{issue.issue_line}: {issue.description}  \033[92m successfully fixed\033[0m"
                )
            else:
                issue_index += 1
                logger.message(
                    f"{issue.filename}:{issue.issue_line}: {issue.description}  \033[91m unable to fix\033[0m"
                )



# Copyright: Hallux team, 2023

from __future__ import annotations

import os
from pathlib import Path
from typing import Final

import openai
from openai.api_resources import ChatCompletion

from hallux.logger import logger

from ..backends.query_backend import QueryBackend
from ..issues.issue import IssueDescriptor


class OpenAiChatGPT(QueryBackend):
    def __init__(
        self,
        model: str = "",
        max_tokens: int = 4097,
        type="openai",
        base_path: Path = Path(),
        previous_backend: QueryBackend | None = None,
    ):
        assert type == "openai" or type == "openai.azure"
        super().__init__(base_path, previous_backend)

        self.valid = True
        if model is None or len(model) < 2:
            logger.warning(f"Wrong model name for OpenAI Backend: {model}")
            self.valid = False

        self.model: Final[str] = model
        self.max_tokens: Final[int] = max_tokens

        # For Azure OpenAI
        if type == "openai.azure":
            openai.api_key = self.validate_env(
                "AZURE_OPENAI_API_KEY", "Environment variable {key} is required for OpenAI Azure Backend"
            )
            openai.api_base = self.validate_env(
                "AZURE_OPENAI_ENDPOINT", "Environment variable {key} is required for OpenAI Azure Backend"
            )
            openai.api_type = "azure"
            openai.api_version = "2023-05-15"
        else:
            openai.api_key = self.validate_env(
                "OPENAI_API_KEY", "Environment variable {key} is required for OpenAI Backend"
            )

    def query(self, request: str, issue: IssueDescriptor | None = None, issue_lines: list[str] = list) -> list[str]:
        if not self.valid:
            return []

        logger.debug("[OpenAI REQUEST]:")
        logger.debug(request)
        result = ChatCompletion.create(messages=[{"role": "user", "content": request}], model=self.model)
        answers = []
        if len(result["choices"]) > 0:
            for variant in result["choices"]:
                answers.append(variant["message"]["content"])

        logger.debug("[OpenAI ANSWERS]:")
        for ans in answers:
            for line in ans.split("\n"):
                logger.debug(line)

        return answers

    def validate_env(self, key: str, message: str) -> None | str:
        if os.getenv(key) is None:
            logger.warning(message.format(key=key))
            self.valid = False
        else:
            return os.getenv(key)



# Copyright: Hallux team, 2023

from __future__ import annotations

import copy
import os
import subprocess

from github import Github, GithubObject, PullRequest, Repository

from ..logger import logger
from ..proposals.diff_proposal import DiffProposal
from .filesystem import FilesystemTarget


# Saves Issue Fixes as Github proposals
class GithubSuggestion(FilesystemTarget):
    def __init__(self, pr_url: str | None):
        FilesystemTarget.__init__(self)

        (base_url, repo_name, PR_ID) = GithubSuggestion.parse_pr_url(pr_url)

        if base_url is None:
            logger.warning(f"Cannot parse github PR URL: {pr_url}")
            raise SystemError(f"Cannot parse github PR URL: {pr_url}")

        if "GITHUB_TOKEN" not in os.environ.keys():
            logger.warning("GITHUB_TOKEN is not provided")
            raise SystemError("GITHUB_TOKEN is not provided")

        self.github = Github(os.environ["GITHUB_TOKEN"], base_url=base_url)
        self.repo: Repository = self.github.get_repo(repo_name)
        self.pull_request: PullRequest = self.repo.get_pull(PR_ID)

        if self.pull_request.closed_at is not None:  # self.pull_request.is_merged :
            raise SystemError(f"Pull Request {PR_ID} is either closed or merged already")

        merge_commit_sha: str = self.pull_request.merge_commit_sha
        latest_github_sha: str = self.pull_request.head.sha
        git_log = subprocess.check_output(["git", "log", "--pretty=oneline"]).decode("utf8")
        local_git_sha = git_log.split("\n")[0].split(" ")[0]

        if not (local_git_sha == latest_github_sha or local_git_sha == merge_commit_sha):
            raise SystemError(
                f"Local git commit: `{local_git_sha}` "
                f"do not coinside with the head from pull-request: `{latest_github_sha}` "
                f"nor with the merge commit: `{merge_commit_sha}`"
            )

    @staticmethod
    def parse_pr_url(pr_url: str) -> tuple[str, str, int] | tuple[None, None, None]:
        """
        Tries to parse Pull-Request URL to obtain base_url, repo_name and PR_ID
        :param pr_url: shall look like this: https://BUSINESS.github.com/YOUR_NAME/REPO_NAME/pull/ID
        :return: (base_url, base_url = "https://api." + url_items[0], PR_ID) OR None
        """
        if pr_url.startswith("https://"):
            pr_url = pr_url[len("https://") :]
            url_items = pr_url.split("/")
            if len(url_items) == 5 and url_items[3] == "pull":
                if url_items[0] == "github.com":
                    base_url = "https://api." + url_items[0]
                else:
                    base_url = "https://api." + url_items[0] + "/api/v3"
                repo_name = url_items[1] + "/" + url_items[2]
                pr_id = int(url_items[4])
                return base_url, repo_name, pr_id
        return None, None, None

    def apply_diff(self, diff: DiffProposal) -> bool:
        for file in self.pull_request.get_files():
            if diff.filename == file.filename:
                FilesystemTarget.apply_diff(self, diff)
                return True
        return False

    def revert_diff(self):
        FilesystemTarget.revert_diff(self)

    @staticmethod
    def compact_proposal(proposal: DiffProposal) -> DiffProposal:
        """
        Tries compacting proposed lines by removing overlapping lines of code
        """

        compacted = copy.deepcopy(proposal)

        # try compacting from the beginning
        prop_len = len(compacted.proposed_lines)
        orig_len = len(compacted.issue_lines)
        for i in range(min(prop_len, orig_len)):
            if compacted.proposed_lines[0] == compacted.issue_lines[i]:
                compacted.proposed_lines = compacted.proposed_lines[1:]
                compacted.start_line += 1
            else:
                break

        # now compacting from the end
        prop_len = len(compacted.proposed_lines)
        for i in range(min(prop_len, orig_len)):
            if compacted.proposed_lines[-1] == compacted.issue_lines[-1 - i]:
                compacted.proposed_lines = compacted.proposed_lines[:-1]
                compacted.end_line -= 1
            else:
                break
        return compacted

    def commit_diff(self) -> bool:
        commit = self.repo.get_commit(self.pull_request.head.sha)
        body = self.existing_proposal.description + "\n```suggestion\n"
        compacted = self.compact_proposal(self.existing_proposal)

        for i in range(len(compacted.proposed_lines)):
            line = compacted.proposed_lines[i]
            body += line
            body += "\n" if i < len(compacted.proposed_lines) - 1 else ""
        body = body + "\n```"

        success: bool = True
        try:
            self.pull_request.create_review_comment(
                body=body,
                commit=commit,
                side="RIGHT",
                path=str(compacted.filename),
                line=compacted.end_line,
                start_line=compacted.start_line if compacted.start_line < compacted.end_line else GithubObject.NotSet,
                start_side="RIGHT" if compacted.start_line < compacted.end_line else GithubObject.NotSet,
            )
        except BaseException as ex:
            logger.debug(ex)
            success = False

        # clean local code
        FilesystemTarget.revert_diff(self)
        return success

    def requires_refresh(self) -> bool:
        return False



# Copyright: Hallux team, 2023

from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Final

from ..issues.issue import IssueDescriptor


class QueryBackend(ABC):
    def __init__(self, base_path: Path = Path(), previous_backend: QueryBackend | None = None):
        self.previous = previous_backend
        self.base_path: Final[Path] = base_path
        self.was_modified = False

    def previous_backend(self) -> QueryBackend | None:
        return self.previous

    @abstractmethod
    def query(self, request: str, issue: IssueDescriptor | None = None, issue_lines: list[str] = list) -> list[str]:
        pass

    def report_succesfull_fix(self, issue, proposal) -> None:
        if self.previous is not None:
            self.previous.report_succesfull_fix(issue, proposal)
